{-# LANGUAGE ForeignFunctionInterface #-}
-- {-# LANGUAGE MultiParamTypeClasses #-}
-- {-# LANGUAGE ScopedTypeVariables #-}

{- | Alternate Enumerator/Iteratee take by David Mazieres.

     An iteratee is a data sink that is fed chunks of data.  It may
     return a useful result, or its use may be in the side-effects it
     has, such as storing the data to a file.  Iteratees are
     represented by the type @'IterF' t m a@.  @t@ is the type of the
     data chunks (which must be a 'Monoid', such as 'String' or lazy
     'ByteString').  @m@ is the 'Monad' in which the iteratee
     runs--for instance 'IO' (or an instance of 'MonadIO') for the
     iteratee to perform IO.  'a' is the result type of the iteratee,
     for when it has consumed enough input to produce a result.

     An Enumerator is a data source that feeds data chunks to an
     iteratee.  There are two types of Enumerator.

       * An /outer enumerator/, represented by the type 'EnumO',
         generates data from some external source, such as IO (or
         potentially some internal state such as a somepseudo-random
         generator).  An outer enumerator use the 'feedO' function to
         feed data chunks to an iteratee.  When the enumerator is out
         of data, it returns the iteratee, so that it can potentially
         be passed to a different enumerator for more data.  (An
         enumerator should not feed 'EOF' to an iteratee--only the
         'run' and 'runI' functions do this.)  If the iteratee returns
         a result, the enumerator also returns it immediately.

       * An /inner enumerator/, represented by the type 'EnumI', gets
         its data from another enumerator, then feeds this to an
         iteratee.  Thus, an 'EnumI' appears as an iteratee to the
         outer enumerator, and apears as an enumerator to some "inner"
         iteratee to which it is feeding data.  Inner enumerators use
         the function 'feedI' to feed data chunks.  An inner
         enumerator, when done, returns the inner iteratee's state, as
         well as its own Iteratee state.  An inner enumerator that
         receives EOF should /not/ feed EOF to its iteratee as the
         iteratee may subsequently be hooked up to another enumerator.
         This is convention is respected by the 'feedI' function,
         which takes data rather than a chunk.

-}

module MiniIter (-- * Base types
                 Chunk(..), IterR(..), IterF(..), EnumO, EnumI
                -- * Core functions
                , run, cat, catI, (|$)
                , (|..), (..|..), (..|)
                , feedO, feedI
                -- * Some basic Iteratees
                , nullI, failI, eofI
                , runI
                , headI, safeHeadI
                , lineI, safeLineI, stringMaxI, stringExactI
                , putI, handleI, sendI
                -- * Some basic Enumerators
                , enumPure, enumHandle, enumDgram, enumDgramFrom
                , enumNop, enumLog, enumhLog, enumSplit, iterLoop
                -- * Utility functions
                , iterRtoF, feed
                , liftIIO
                , recvStrFrom, sendStr, hShutdown
                ) where

import Control.Concurrent.MVar
import Control.Exception (handle, try, throw, onException)
import Control.Monad
import Control.Monad.Fix
import Control.Monad.Trans
import Data.Int
import Data.Monoid
-- import Data.Word
import Foreign.C
import Foreign.Ptr
import qualified Data.ByteString as S
import qualified Data.ByteString.Internal as S
import qualified Data.ByteString.Unsafe as S
import qualified Data.ByteString.Lazy.Internal as L (defaultChunkSize)
import qualified Data.ByteString.Lazy.Char8 as L8
import qualified Data.ByteString.Lazy as L
import Network.Socket
import System.IO
import System.IO.Error (isEOFError)

import GHC.IOBase (Handle(..), Handle__(..))

-- import Debug.Trace

foreign import ccall unsafe "sys/socket.h send"
  c_send :: CInt -> Ptr a -> CSize -> CInt -> IO CInt
foreign import ccall unsafe "sys/socket.h shutdown"
  c_shutdown :: CInt -> CInt -> IO CInt


--
-- Iteratee types
--

-- | A chunk of data
data Chunk t = Chunk t
             | EOF
               deriving (Eq, Show)

instance (Monoid t) => Monoid (Chunk t) where
    mempty                      = Chunk mempty
    mappend (Chunk a) (Chunk b) = Chunk $ mappend a b
    mappend EOF _               = error "wrote to Iteratee after EOF"
    mappend c EOF               = c

-- XXX should probably have separate IterFail and EnumFail cases,
-- because if the Enumerator fails you may still want the state of the
-- Iteratee.

-- | An Iteratee error
data IterE = IterFail String    -- ^ Iteratee monadic fail function called
           | IOFail IOError     -- ^ IO error that occured

-- | The result of iterating over input
data IterR t m a = Done a (Chunk t)
                 | Cont (IterF t m a)
                 | Fail IterE

-- | A function to iterate over input
newtype IterF t m a = IterF { runIter :: Chunk t -> m (IterR t m a) }

instance (Monad m) => Functor (IterF t m) where
    fmap = liftM

instance (Monad m) => Monad (IterF t m) where
    return a = IterF $ \c -> return $ Done a c
    i >>= k  = IterF $ \c ->
               do result <- runIter i c
                  case result of
                    Cont i'   -> return $ Cont $ i' >>= k
                    Done a c' -> runIter (k a) c'
                    Fail err  -> return $ Fail err
    fail msg = failI $ IterFail msg

instance (MonadFix m) => MonadFix (IterF t m) where
    mfix f = IterF $ \c -> mfix $ \ ~(Done a _) -> runIter (f a) c

instance MonadTrans (IterF t) where
    lift m = IterF $ \c -> m >>= return . flip Done c

instance (MonadIO m) => MonadIO (IterF t m) where
    liftIO = lift . liftIO

-- | Lift an IO operation into an IterF monad, but if the IO operation
-- throws an error, catch the exception and return it as a failure of
-- the Iteratee.
liftIIO :: (MonadIO m) => IO a -> IterF t m a
liftIIO m = do
  result <- liftIO $ try m
  case result of
    Left err -> failI $ IOFail err
    Right ok -> return ok

-- | Turn the result of an Iteratee back into an Iteratee.
iterRtoF :: (Monad m, Monoid t) => IterR t m a -> IterF t m a
iterRtoF (Cont f)   = f
iterRtoF (Fail e)   = failI e
iterRtoF (Done a c) = IterF $ \c' -> return $ Done a $ mappend c c'

-- | Feed EOF to an iteratee, and return the result
run :: (Monad m) => IterF t m a -> m a
run iter = runIter iter EOF >>= check
    where
      check (Done a _)          = return a
      check (Cont _)            = error "IterF returned Cont after EOF"
      check (Fail (IOFail e))   = throw e
      check (Fail (IterFail e)) = fail e

--
-- Some super-basic Iteratees
--

-- | Sinks data like /dev/null and only returns on EOF
nullI :: (Monad m, Monoid t) => IterF t m ()
nullI = IterF $ return . check
    where
      check EOF = Done () EOF
      check _   = Cont nullI

-- | Iteratee that always fails
failI :: (Monad m) => IterE -> IterF t m a
failI err = IterF $ \_ -> return $ Fail err

{-
-- | Iteratee that just returns the next non-empty chunk
chunkI :: (Monad m, Monoid t, Eq t) => IterF t m (Chunk t)
chunkI = IterF $ return . check
    where
      check EOF                     = Done EOF EOF
      check (Chunk c) | c == mempty = Cont chunkI
      check c                       = Done c $ Chunk mempty
-}

-- | Returns true at end of file
eofI :: (Monad m, Monoid t) => IterF t m Bool
eofI = IterF $ return . check
    where
      check EOF = Done True $ EOF
      check c   = Done False c

-- | Runs an Iteratee from within another Iteratee.  This is different
-- from just extracting the value on the next chunk of input, as runI
-- feeds an EOF to its iteratee.
runI :: (Monad m) => IterF tIn m a -> IterF tOut m a
runI iter = lift $ run iter

-- | Return the the first element when the Iteratee data type is a list.
headI :: (Monad m) => IterF [a] m a
headI = IterF $ return . dohead
    where
      dohead EOF = Fail $ IterFail "headI: encountered EOF"
      dohead (Chunk []) = Cont headI
      dohead (Chunk (a:as)) = Done a $ Chunk as

-- | Return the the first element when the Iteratee data type is a list.
safeHeadI :: (Monad m) => IterF [a] m (Maybe a)
safeHeadI = IterF $ return . dohead
    where
      dohead EOF = Done Nothing EOF
      dohead (Chunk []) = Cont safeHeadI
      dohead (Chunk (a:as)) = Done (Just a) $ Chunk as

-- | Return a line delimited by \r, \n, or \r\n.
safeLineI :: (Monad m) => IterF L.ByteString m (Maybe L.ByteString)
safeLineI = IterF $ return . doline L.empty
    where
      eol c = c == '\r' || c == '\n'
      doline acc EOF =
          let (l, r) = L8.break eol acc
          in if L.null r then Done Nothing (Chunk acc) else Done (Just l) EOF
      doline acc (Chunk t) =
          let acc' = L.append acc t
              (l, r) = L8.break eol acc'
          in case () of
               () | L.null r || r == L8.singleton '\r' ->
                      Cont $ IterF $ return . doline acc'
                  | L.take 2 r == L8.pack "\r\n" ->
                      Done (Just l) $ Chunk $ L.drop 2 r
                  | otherwise ->
                      Done (Just l) $ Chunk $ L.drop 1 r

-- | Return a line delimited by \r, \n, or \r\n.
lineI :: (Monad m) => IterF L.ByteString m L.ByteString
lineI = do
  mline <- safeLineI
  case mline of
    Nothing -> fail "lineI: incomplete line"
    Just line -> return line

-- | Return a string that is at most maxlen bytes, and at least one
-- byte unless EOF is encountered, in which case the null string is
-- returned.
stringMaxI :: (Monad m) => Int64 -> IterF L.ByteString m L.ByteString
stringMaxI maxlen = IterF $ return . dostring
    where
      dostring EOF       = Done L.empty EOF
      dostring (Chunk s) = if L.null s && maxlen > 0
                             then Cont $ stringMaxI maxlen
                             else case L.splitAt maxlen s of
                                    (h, t) -> Done h $ Chunk t

-- | Return a sring that is exactly len bytes unless an EOF is
-- encountered.
stringExactI :: (Monad m) => Int64 -> IterF L.ByteString m L.ByteString
stringExactI len | len <= 0  = return L.empty
                 | otherwise = accumulate L.empty
    where
      accumulate acc = do
        t <- stringMaxI (len - L.length acc)
        if L.null t then return acc else
            let acc' = L.append acc t
            in if L.length t == len then return acc' else accumulate acc'

-- | An Iteratee that puts byte strings to a consumer function, then
-- call an eof function.  For instance, 'handleI' is defined as:
--
-- > handleI :: (MonadIO m) => Handle -> IterF L.ByteString m ()
-- > handleI h = putI (liftIIO . L.hPut h) (liftIIO $ hShutdown h 1)
--
putI :: (Monad m) =>
        (L.ByteString -> IterF L.ByteString m a)
     -> (IterF L.ByteString m b)
     -> IterF L.ByteString m ()
putI putfn eoffn = do
  buf <- stringMaxI $ fromIntegral L.defaultChunkSize
  if L.null buf then eoffn >> return () else putfn buf >> putI putfn eoffn

-- | Put byte strings to a file handle then close it
handleI :: (MonadIO m) => Handle -> IterF L.ByteString m ()
handleI h = putI (liftIIO . L.hPut h) (liftIIO $ hShutdown h 1)

-- | Send datagrams using a supplied function
sendI :: (Monad m) =>
         (L.ByteString -> IterF [L.ByteString] m a)
      -> IterF [L.ByteString] m ()
sendI sendfn = do
  dgram <- safeHeadI
  case dgram of
    Just pkt -> sendfn pkt >> sendI sendfn
    Nothing  -> return ()

--
-- Enumerator types
--

-- | An outermost enumerator that gets data from somewhere else and
-- feeds it to the iteratee, returning the iteratee when there is no
-- more input or the iteratee produces a result of type a.
type EnumO t m a = IterF t m a -> IterF t m a

-- | Concatenate two outer enumerators, forcing them to be executed in
-- turn in the monad @m@.  Note that the deceptively simple definition:
--
--  >  cat a b = b . a
--
-- wouldn't necessarily do the right thing, as in this case @a@ would
-- not actually get to run until @b@ executes it, and @b@ might start
-- out, before feeding any input to its enumerator, by waiting for
-- some event that is triggered by a side-effect of @a@.
cat :: (Monad m, Monoid t) => EnumO t m a -> EnumO t m a -> EnumO t m a
cat a b iter = IterF iterf
    where
      iterf c = do
        r <- runIter (a iter) $ Chunk mempty
        case r of
          Cont f -> runIter (b f) c
          _      -> return r
infixr 4 `cat`

-- | Run an outer enumerator on an iteratee
(|$) :: (Monad m) => EnumO t m a -> IterF t m a -> m a
(|$) enum iter = run (enum iter)
infix 2 |$

-- | An inner enumerator or transcoder.  Such a function accepts data
-- from some outer enumerator (acting like an Iteratee), then
-- transcodes the data and feeds it to another IterF (hence also
-- acting like an enumerator towards that inner IterF).  Note that
-- data is viewed as flowing inwards from the outermost enumerator to
-- the innermost iteratee.  Thus tOut, the "outer type" is actually
-- the type of input fed to an EnumI, while tIn what the EnumI outputs
-- by feeding to the iteratee.
type EnumI tOut tIn m a = IterF tIn m a -> IterF tOut m (IterF tIn m a)

-- | Concatenate two inner enumerators
catI :: (Monad m) => EnumI tOut tIn m a -> EnumI tOut tIn m a
     -> EnumI tOut tIn m a
catI a b iter = a iter >>= b
infixr 4 `catI`

-- | Compose an outer enumerator, producing chunks of some type
-- @tOut@, with an inner enumerator that transcodes @tOut@ to @tIn@,
-- to produce a new outer enumerator producing chunks of type @tIn@.
(|..) :: (Monad m) =>
         EnumO tOut m (IterF tIn m a)
      -> EnumI tOut tIn m a
      -> EnumO tIn m a
(|..) outer inner iter = lift (run $ outer $ inner iter) >>= id
infixr 3 |..

{- The above function is maybe a bit easier to understand if you
   follow the types.  With ScopedTypeVariables, the types work out as
   follows:

    let ii  = inner iter :: IterF tOut m (IterF tIn m a)
        oii = outer ii   :: IterF tOut m (IterF tIn m a)
        r   = run oii    :: m (IterF tIn m a)
        l   = lift r     :: IterF tIn m (IterF tIn m a)
    in l >>= id
-}

-- | Fuse two inner enumerators into one.
(..|..) :: (Monad m) => 
           EnumI tOut tMid m (IterF tIn m a)
        -> EnumI tMid tIn m a
        -> EnumI tOut tIn m a
(..|..) outer inner iter = outer (inner iter) >>= runI
infixr 3 ..|..

-- | Fuse an inner enumerator that transcodes @tOut@ to @tIn@ with an
-- iteratee taking type @tIn@ to produce an iteratee taking type
-- @tOut@.
(..|) :: (Monad m) =>
         EnumI tOut tIn m a
      -> IterF tIn m a
      -> IterF tOut m a
(..|) inner iter = inner iter >>= runI
infixr 3 ..|

-- | Feed pure data directly to an iteratee.
feed :: (Monad m, Monoid t) => t -> IterF t m a -> m (IterR t m a)
feed t iter = runIter iter $ Chunk t

-- | Feed pure data directly to an iteratee from within an EnumO
-- function.  Takes the outer EnumO as an argument and invokes it
-- recursively when the iteratee returns Cont.
feedO :: (Monad m, Monoid t) =>
         EnumO t m a            -- ^ Outer Enumerator (that gets more data)
      -> t                      -- ^ Data to feed to the Iteratee
      -> EnumO t m a            -- ^ Takes an IterF and feeds it the data
feedO enum t iter = do
  result <- lift $ feed t iter
  case result of
    Cont iter' -> enum iter'
    done       -> iterRtoF done

-- | Feed a chunk of the inner data type to the inner iteratee from
-- within an EnumI.  Calls the first EnumI argument recursively when
-- the inner Iteratee is not done.
feedI :: (Monad m, Monoid tIn) =>
         EnumI tOut tIn m a     -- ^ Outer Enumerator
      -> tIn                    -- ^ Transcoded data to feed to inner IterF
      -> EnumI tOut tIn m a     -- ^ Takes the inner IterF and feeds it the data
feedI enum t iter = do
    result <- lift $ feed t iter
    case result of
      Cont iter' -> enum iter'
      done       -> return $ iterRtoF done

--
-- Basic outer enumerators
--

-- | Enumerator that feeds pure data to iterators
enumPure :: (Monad m, Monoid t) => t -> EnumO t m a
enumPure t iter = lift (feed t iter) >>= iterRtoF

-- | Read datagrams from a socket and feed a list of strings (one for
-- each datagram) into an Iteratee.
enumDgram :: (MonadIO m) => Socket
          -> EnumO [L.ByteString] m a
enumDgram sock iter = do
  (msg, r, _) <- liftIIO $ recvStrFrom sock 0x10000
  if r < 0 then iter else feedO (enumDgram sock) [msg] iter

-- | Read datagrams from a socket and feed a list of (Bytestring,
-- SockAddr) pairs (one for each datagram) into an Iteratee.
enumDgramFrom :: (MonadIO m) => Socket
          -> EnumO [(L.ByteString, SockAddr)] m a
enumDgramFrom sock iter = do
  (msg, r, addr) <- liftIIO $ recvStrFrom sock 0x10000
  if r < 0 then iter else feedO (enumDgramFrom sock) [(msg, addr)] iter

-- | Feed data from a file handle into an Iteratee.
enumHandle :: (MonadIO m) => Handle
              -> EnumO L.ByteString m a
enumHandle h iter = do
  -- liftIO $ putTraceMsg "about to wait for IO"
  buf <- liftIIO $ handle eofHandler $
         hWaitForInput h (-1) >> L.hGetNonBlocking h L.defaultChunkSize
  -- liftIO $ putTraceMsg $ "just got " ++ (show $ L.length buf) ++ " bytes"
  if L.null buf then iter else feedO (enumHandle h) buf iter
    where
      eofHandler e = if isEOFError e then return L.empty else ioError e

-- | Create a loop-back (iteratee, enumerator) pair
iterLoop :: (MonadIO m, Monoid t, Show t) =>
            m (IterF t m (), EnumO t m a)
iterLoop = do
  pipe <- liftIO $ newEmptyMVar >>= newMVar
  return (IterF $ iterf pipe, enum pipe)
    where
      iterf pipe EOF = do
             liftIO $ withMVar pipe $ \p ->
                 do mp <- tryTakeMVar p
                    putMVar p $ case mp of
                       Just (_, c) -> (True, c)
                       Nothing     -> (True, EOF)
             return $ Done () mempty
      iterf pipe c = do
             liftIO $ withMVar pipe $ \p ->
                 do mp <- tryTakeMVar p
                    putMVar p $ case mp of
                       Nothing          -> (False, c)
                       Just (True, _)   ->
                           error $ "iterLoop: write after EOF (" ++ show c ++ ")"
                       Just (False, c') -> (False, mappend c' c)
             return $ Cont $ IterF $ iterf pipe

      enum pipe iter = do
             p <- liftIO $ readMVar pipe
             (eof, c) <- liftIO $ takeMVar p
             when eof $ iterf pipe EOF >> return ()
             case c of
               Chunk t -> feedO (enum pipe) t iter
               EOF     -> iter

--
-- Basic inner enumerators
--

-- | The null inner enumerator.  Passes input through to another
-- iteratee.  On EOF returns the iteratee (without passing through the
-- EOF).  Also returns if the inter iteratee returns a value or error.
enumNop :: (Monad m, Monoid t) => EnumI t t m a
enumNop iter@(IterF f) = IterF iterf
    where
      iterf EOF = return $ Done iter mempty
      iterf c   = do
        r <- f c
        return $ case r of
                   Cont iter' -> Cont $ enumNop iter'
                   done       -> Done (iterRtoF done) mempty

{-
enumLog :: (MonadIO m) =>
           FilePath             -- ^ Path to log to
        -> Bool                 -- ^ True to truncate file
        -> EnumI L.ByteString L.ByteString m a
enumLog path trunc iter1 = do
  h <- liftIIO $ openFile path (if trunc then AppendMode else WriteMode)
  liftIIO $ hSetBuffering h NoBuffering
  enum h iter1
    where
      enum h iter = do
        buf <- stringMaxI $ fromIntegral L.defaultChunkSize
        if L.null buf
          then liftIIO (hClose h) >> return iter
          else do liftIIO $ L.hPut h buf `onException` hClose h
                  feedI (enum h) buf iter
-}

enumLog :: (MonadIO m) =>
           FilePath             -- ^ Path to log to
        -> Bool                 -- ^ True to truncate file
        -> EnumI L.ByteString L.ByteString m a
enumLog path trunc iter = do
  h <- liftIIO $ openFile path (if trunc then WriteMode else AppendMode)
  liftIIO $ hSetBuffering h NoBuffering
  enumhLog h iter
          
enumhLog :: (MonadIO m) =>
            Handle
         -> EnumI L.ByteString L.ByteString m a
enumhLog h iter = do
  buf <- stringMaxI $ fromIntegral L.defaultChunkSize
  if L.null buf
    then liftIIO (hClose h) >> return iter
    else do liftIIO $ L.hPut h buf `onException` hClose h
            feedI (enumhLog h) buf iter
          

-- | Returns an Iteratee that can be written from multiple threads, as
-- long as the inner iteratee does not throw an exception (e.g.,
-- should use 'liftIIO' instead of 'liftIO').
enumSplit :: (MonadIO m, Monoid t) => EnumI t t m a
enumSplit iter1 = do
  mv <- liftIO $ newMVar $ Cont iter1
  IterF $ iterf mv
    where
      iterf mv EOF = do
        r <- liftIO $ readMVar mv
        return $ Done (iterRtoF r) mempty
      iterf mv c = do
        rold <- liftIO $ takeMVar mv
        rnew <- case rold of
                  Cont (IterF f)      -> f c
                  Done a c'@(Chunk _) -> return $ Done a $ mappend c' c
                  other               -> return other
        liftIO $ putMVar mv rnew
        return $ case rnew of
                   Cont _    -> Cont $ IterF $ iterf mv
{-
                   -- Pull left-over data from the inner iteratee
                   -- result to the outer iteratee result, so that any
                   -- left-over data will be returned by (..|)
                   Done a c' -> Done (return a) c'
-}
                   _         -> Done (iterRtoF rnew) mempty


--
-- Some utility functions for things that are made hard by the Haskell
-- libraries
--

-- | Receive data from a datagram socket into a lazy bytestring.
recvStrFrom :: Socket -> Int -> IO (L.ByteString, Int, SockAddr)
recvStrFrom s len = do
  (str, (r, addr)) <- S.createAndTrim' (max 0 len) callRecv
  return (L.fromChunks [str], r, addr)
    where
      callRecv ptr = do (r, addr) <- recvBufFrom s ptr len
                        return (0, max r 0, (r, addr))

sendStr :: Socket -> L.ByteString -> IO Int
sendStr (MkSocket s _ _ _ _) str = do
  r <- S.unsafeUseAsCStringLen (S.concat $ L.toChunks str) $
       \(p, n) -> c_send s p (fromIntegral n) 0
  return $ fromIntegral r

hShutdown                            :: Handle -> CInt -> IO Int
hShutdown h@(FileHandle _ m) how     = hFlush h >> hShutdown' m how
hShutdown h@(DuplexHandle _ _ m) how = hFlush h >> hShutdown' m how
hShutdown'       :: MVar Handle__ -> CInt -> IO Int
hShutdown' m how = do -- putTraceMsg "calling hShutdown'"
                      withMVar m $ \(Handle__ { haFD = fd }) ->
                          liftM fromIntegral $ c_shutdown (fromIntegral fd) how
  
